#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/file.h>

#include <postgres.h>
#include <fmgr.h>
#include <commands/trigger.h>
#include <executor/spi.h>

PG_MODULE_MAGIC;

extern void set_ps_display (const char *, bool);

#define INCARNATION "SYSTEM.INCARNATION"

#define SETTINGS_QUERY(key, append_value, notfound_value) \
  "(SELECT CASE " \
  "WHEN setting IS NULL OR char_length(setting) = 0 " \
  "THEN '" notfound_value "' " \
  "ELSE setting || '" append_value "' END " \
  "FROM pg_settings WHERE name = '" key "')"

#define PID_FILE_QUERY SETTINGS_QUERY("data_directory", "/postmaster.pid", "postmaster.pid")

#define SOCKET_QUERY \
  SETTINGS_QUERY("unix_socket_directory", "", "/tmp") \
  " || '/.s.PGSQL.' || " \
  SETTINGS_QUERY("port", "", "5432")

#define FILE_QUERY(a) SETTINGS_QUERY(#a "_file", "", "pg_" #a ".conf")

#define DISCLAIMER_QUERY(a) \
  "'## This file has automatically been created. Do NOT modify it manually!\n" \
  "## Modify the corresponding system table system." #a " instead.\n\n'"

static void spi_connect ()
{
  if (SPI_connect() != SPI_OK_CONNECT)
    elog(ERROR, "%s", "cannot connect to SPI");
}

static void spi_disconnect ()
{
  if (SPI_finish() != SPI_OK_FINISH)
    elog(ERROR, "%s", "cannot disconnect from SPI");
}

static char *get_pid_file ()
{
  if (SPI_execute("SELECT " PID_FILE_QUERY " AS a",
                  false, 1) != SPI_OK_SELECT)
    elog(ERROR, "%s", "can not get pidfile");

  return SPI_getvalue(SPI_tuptable->vals[0], SPI_tuptable->tupdesc, 1);
}

static pid_t get_master_pid (char *pidfile)
{
  int fh, i;
  ssize_t len;
  char buf[64];
  pid_t pid;

  if ((fh = open(pidfile, O_RDONLY)) < 0) return -1;
  if ((len = read(fh, buf, 64)) < 0) i = errno;
  else i = 0;
  close(fh);
  if (i) {
    errno = i;
    return -1;
  }

  for (i = 0, pid = 0; i < len; ++i) {
    if ((buf[i] < '0') || (buf[i] > '9')) break;
    pid = pid * 10 + buf[i] - '0';
  }

  return pid;
}

static void notify_postgres (int sig)
{
  pid_t pid;
  char *pidfile;

  spi_connect();
  pidfile = get_pid_file();
  spi_disconnect();

  pid = get_master_pid(pidfile);
  pfree(pidfile);

  if (pid < 0)
    elog(ERROR, "can not read %s: %s", pidfile, strerror(errno));

  if (kill(pid, sig))
    elog(ERROR, "can not notify postmaster (pid %i): %s", pid, strerror(errno));
}

static void restart_postgres (int sig)
{
  pid_t pid;
  int i;
  char *c, buf[64], *oldpath, *newpath, *pidfile;
  long count, count_save;
  struct stat st;

  spi_connect();
  if (SPI_execute("SELECT " SOCKET_QUERY " AS a, "
                  SOCKET_QUERY " || '.restart' AS b",
                  false, 1) != SPI_OK_SELECT)
    elog(ERROR, "%s", "can not determine socket path");
  oldpath = SPI_getvalue(SPI_tuptable->vals[0], SPI_tuptable->tupdesc, 1);
  newpath = SPI_getvalue(SPI_tuptable->vals[0], SPI_tuptable->tupdesc, 2);

  pidfile = get_pid_file();

  if (rename(oldpath, newpath)) {
    if (stat(newpath, &st))
      elog(ERROR, "shutdown failed: %s", "can not find socket");
    elog(ERROR, "shutdown failed: %s", "shutdown already in progress");
  }

  spi_disconnect();

  if ((pid = fork()) < 0)
    elog(ERROR, "can not fork: %s", strerror(errno));
 
  if (pid > 0) {
    notify_postgres(sig);
    return;
  }

  set_ps_display("restarting...", true);

  for (i = 0; i < 1024; ++i) close(i);
  setsid();

  strcpy(buf, INCARNATION);
  strcat(buf, "=");
  if ((c = getenv(INCARNATION))) {
    count = atol(c) + 1;
    count_save = count;
    c = buf + strlen(buf);
    do {
      ++c;
      count /= 10;
    } while (count);
    *c = '\0';
    do {
      *--c = count_save % 10 + '0';
      count_save /= 10;
    } while (count_save);
  }
  else strcat(buf, "0");

  for (;;) {
    pid = get_master_pid(pidfile);
    if (pid < 0) break;
    if (kill(pid, 0)) break;
    sleep(1);
  }

  if ((i = open("/dev/zero", O_RDONLY))) {
    dup2(i, 0);
    close(i);
  }
  if ((i = open("/dev/null", O_WRONLY)) != 1) {
    dup2(i, 1);
    close(i);
  }
  dup2(1, 2);

  execve("/usr/bin/postgres", (char*[]) { "/usr/bin/postgres", "-D", ".", 0 },
         (char*[]) { buf, 0 });

  exit(1);
}

PG_FUNCTION_INFO_V1(incarnation);
Datum incarnation (PG_FUNCTION_ARGS)
{
  char *c;

  fcinfo = fcinfo;
  if (!(c = getenv(INCARNATION))) c = "0";
  PG_RETURN_CSTRING(c);
}

PG_FUNCTION_INFO_V1(smart_restart);
Datum smart_restart (PG_FUNCTION_ARGS)
{
  const char *msg = "Database smart restart requested. "
                    "Please log off now.";
  fcinfo = fcinfo;
  restart_postgres(SIGTERM);
  PG_RETURN_CSTRING(msg);
}

PG_FUNCTION_INFO_V1(fast_restart);
Datum fast_restart (PG_FUNCTION_ARGS)
{
  const char *msg = "Database fast restart requested. "
                    "Please reconnect now.";
  fcinfo = fcinfo;
  restart_postgres(SIGINT);
  elog(FATAL, "%s", msg);
  PG_RETURN_CSTRING(msg);
}

PG_FUNCTION_INFO_V1(immediate_restart);
Datum immediate_restart (PG_FUNCTION_ARGS)
{
  const char *msg = "Database immediate restart requested. "
                    "Please reconnect now.";
  fcinfo = fcinfo;
  restart_postgres(SIGQUIT);
  elog(FATAL, "%s", msg);
  PG_RETURN_CSTRING(msg);
}

PG_FUNCTION_INFO_V1(reload);
Datum reload (PG_FUNCTION_ARGS)
{
  fcinfo = fcinfo;
  notify_postgres(SIGHUP);
  PG_RETURN_CSTRING("Database reload requested.");
}

PG_FUNCTION_INFO_V1(smart_shutdown);
Datum smart_shutdown (PG_FUNCTION_ARGS)
{
  const char *msg = "Database smart shutdown requested. "
                    "Please disconnect now.";
  fcinfo = fcinfo;
  notify_postgres(SIGTERM);
  PG_RETURN_CSTRING(msg);
}

PG_FUNCTION_INFO_V1(fast_shutdown);
Datum fast_shutdown (PG_FUNCTION_ARGS)
{
  const char *msg = "Database fast shutdown requested. "
                    "Your session has been disconnected.";
  fcinfo = fcinfo;
  notify_postgres(SIGINT);
  elog(FATAL, "%s", msg);
  PG_RETURN_CSTRING(msg);
}

PG_FUNCTION_INFO_V1(immediate_shutdown);
Datum immediate_shutdown (PG_FUNCTION_ARGS)
{
  const char *msg = "Database immediate shutdown requested. "
                    "Your session has been disconnected.";
  fcinfo = fcinfo;
  notify_postgres(SIGQUIT);
  elog(FATAL, "%s", msg);
  PG_RETURN_CSTRING(msg);
}

PG_FUNCTION_INFO_V1(update_connect_file);
Datum update_connect_file (PG_FUNCTION_ARGS)
{
  TriggerData *trigdata = NULL;
  HeapTuple heaptuple = NULL;
  Relation relation = NULL;
  TupleDesc tupledesc = NULL;
  int fh;
  char *relname, *sql = NULL, *value, *dstfile, *tmpfile, *disclaimer;

  if (!fcinfo) elog(ERROR, "%s is NULL", "fcinfo");
  if (!CALLED_AS_TRIGGER(fcinfo)) elog(ERROR, "%s", "not called as trigger");
  if (!(trigdata = (TriggerData*) fcinfo->context))
    elog(ERROR, "%s is NULL", "trigdata");
  if (!TRIGGER_FIRED_FOR_STATEMENT(trigdata->tg_event))
    elog(ERROR, "not called %s", "for statement");
  if (!TRIGGER_FIRED_AFTER(trigdata->tg_event))
    elog(ERROR, "not called %s", "after event");
  if (!(relation = trigdata->tg_relation))
    elog(ERROR, "%s is NULL", "relation");
  if (!(tupledesc = relation->rd_att)) elog(ERROR, "%s is NULL", "tupledesc");
  if (TRIGGER_FIRED_BY_UPDATE(trigdata->tg_event))
    heaptuple = trigdata->tg_newtuple;
  else heaptuple = trigdata->tg_trigtuple;

  if (!(relname = RelationGetRelationName(relation)))
    elog(ERROR, "%s is NULL", "relation");
  if (!strcmp(relname, "hba"))
    sql = "SELECT " DISCLAIMER_QUERY(hba) " AS a, "
          "(SELECT array_to_string(ARRAY(SELECT \"type\" || E'\\t' || "
          "\"database\" || E'\\t' || \"user\" || E'\\t' || "
          "CASE WHEN \"type\" = 'local' THEN \'\' ELSE \"host\"::TEXT END || "
          "E'\\t' || \"method\" || E'\\t' || \"option\" || E'\\n' "
          "FROM system.hba), '')) AS b, "
          FILE_QUERY(hba) " AS c, " FILE_QUERY(hba) " || '.tmp' AS d";
  if (!strcmp(relname, "ident"))
    sql = "SELECT " DISCLAIMER_QUERY(ident) " AS a, "
          "(SELECT array_to_string(ARRAY(SELECT \"map\" || E'\\t' || "
          "\"os_user\" || E'\\t' || \"pg_user\" || E'\\n' "
          "FROM system.ident), '')) AS b, "
          FILE_QUERY(ident) " AS c, " FILE_QUERY(ident) " || '.tmp' AS d";
  if (!sql) elog(ERROR, "unknown table %s", relname);

  if (SPI_execute(sql, false, 1) != SPI_OK_SELECT)
    elog(ERROR, "can not get value from table %s", relname);
  if (!(dstfile = SPI_getvalue(SPI_tuptable->vals[0], SPI_tuptable->tupdesc, 3)))
    elog(ERROR, "can not get %s filename", relname);
  if (!(tmpfile = SPI_getvalue(SPI_tuptable->vals[0], SPI_tuptable->tupdesc, 4)))
    elog(ERROR, "can not get %s temp filename", relname);
  value = SPI_getvalue(SPI_tuptable->vals[0], SPI_tuptable->tupdesc, 2);
  disclaimer = SPI_getvalue(SPI_tuptable->vals[0], SPI_tuptable->tupdesc, 1);

  fh = open(tmpfile, O_CREAT | O_TRUNC | O_WRONLY, S_IRUSR | S_IWUSR);
  if (fh < 0) elog(ERROR, "can not write file %s", tmpfile);
  write(fh, disclaimer, strlen(disclaimer));
  if (value) {
    write(fh, value, strlen(value));
    pfree(value);
  }
  close(fh);
  if (rename(tmpfile, dstfile))
    elog(ERROR, "can not rename file %s to file %s", tmpfile, dstfile);
  pfree(disclaimer);
  pfree(tmpfile);
  pfree(dstfile);

  notify_postgres(SIGHUP);

  PG_RETURN_POINTER(heaptuple);
}
